"""
Trajan API
"""
import collections
import numpy as np
import xarray as xr

from . import trajectory_accessor as _
from . import skill as _

def trajectory_dict_to_dataset(trajectory_dict, variable_attributes={}, global_attributes={}):
    """Create a CF-compatible trajectory file from dictionary of drifter positions

    Trajectory_dict shall have the following structure:
        {'buoy1_name': {
            time0: {'lon': lon0, 'lat': lat0, 'variable1': var1_0, ... 'variableM': varM_0},
            time1: {'lon': lon1, 'lat': lat1, 'variable1': var1_1, ... 'variableM': varM_1},
                ...
            timeN: {'lon': lonN, 'lat': latN, 'variable1': var1_N, ... 'variableM': varM_N}},
        {'buoy2_name': {
            ...
    """

    drifter_names = list(trajectory_dict)
    num_drifters = len(drifter_names)
    num_times = np.max([len(d) for dn, d in trajectory_dict.items()])
    # Allocate  arrays
    variables = []
    for dn, d in trajectory_dict.items():
        v = list(d[list(d)[0]])
        for var in v:
            if var not in variables:
                variables.extend([var])
    datavars = {}
    for var in variables:
        datavars[var] = np.empty((num_drifters, num_times))
        datavars[var][:] = np.nan
        if var not in variable_attributes:
            if var == 'lat':
                variable_attributes[var] = {'standard_name': 'latitude', 'unit': 'degree_north'}
            elif var == 'lon':
                variable_attributes[var] = {'standard_name': 'longitude', 'unit': 'degree_east'}
            else:
                variable_attributes[var] = {}
    time = np.empty((num_drifters, num_times), dtype='datetime64[s]')
    time[:] = np.datetime64('nat')

    # Fill arrays with data from dictionaries
    for drifter_num, (drifter_name, drifter_dict) in enumerate(trajectory_dict.items()):
        # Make sure that positions are sorted by time
        drifter_dict = collections.OrderedDict(sorted(drifter_dict.items()))
        t = slice(0, len(drifter_dict))
        for var in variables:
            datavars[var][drifter_num, t] = np.array([di[var] for d, di in drifter_dict.items()])
        time[drifter_num, t] = np.array(list(drifter_dict), dtype='datetime64[s]')

    datavars = {v: (['trajectory', 'obs'], datavars[v], variable_attributes[v]) for v in variables}
    datavars['time'] = (['trajectory', 'obs'], time, {'standard_name': 'time'})
    datavars['drifter_names'] = (['trajectory'], drifter_names,
                {'cf_role': 'trajectory_id', 'standard_name': 'platform_id'})

    # Remove empty attributes
    global_attributes = {a:v for a,v in global_attributes.items() if v != ''}

    # Create Xarray Dataset adhering to CF conventions h.4.1 for trajectory data
    ds = xr.Dataset(
        data_vars=datavars,
        attrs={'Conventions': 'CF-1.10',
               'featureType': 'trajectory',
               'geospatial_lat_min': np.nanmin(datavars['lat'][1]),
               'geospatial_lat_max': np.nanmax(datavars['lat'][1]),
               'geospatial_lon_min': np.nanmin(datavars['lon'][1]),
               'geospatial_lon_max': np.nanmax(datavars['lon'][1]),
               'time_coverage_start': str(np.nanmin(datavars['time'][1])),
               'time_coverage_end': str(np.nanmax(datavars['time'][1])),
                **global_attributes
               }
    )

    return ds



"""
Extending xarray Dataset with functionality specific to trajectory datasets.

Presently supporting Cf convention H.4.1
https://cfconventions.org/Data/cf-conventions/cf-conventions-1.10/cf-conventions.html#_multidimensional_array_representation_of_trajectories.
"""

import numpy as np
from scipy.interpolate import interp1d
import pandas as pd
import xarray as xr
import trajan as ta
import logging

from .plot import Plot

logger = logging.getLogger(__name__)

@xr.register_dataset_accessor("traj")
class TrajAccessor:
    _obj = None
    __plot__ = None

    def __init__(self, xarray_obj):
        self._obj = xarray_obj
        
        # Identify dimension names
        for dim in ['obs', 'time']:
            if dim in self._obj.dims:
                self.timedim = dim
        for dim in ['traj', 'trajectory']:
            if dim in self._obj.dims:
                self.trajdim = dim
        if not hasattr(self, 'timedim'):
            raise ValueError(f'Time dimension not identified: {self._obj.dims}')

    @property
    def plot(self):
        if self.__plot__ is None:
            logger.debug(f'Setting up new plot object.')
            self.__plot__ = Plot(self._obj)

        return self.__plot__

    def gridtime(self, times):
        """Interpolate dataset to regular time interval

        times:
            - an array of times, or
            - a string "freq" specifying a Pandas daterange (e.g. 'h', '6h, 'D'...)

        Note that the resulting DataSet will have "time" as a dimension coordinate.
        """

        if isinstance(times, str):  # Make time series with given interval
            freq = times
            start_time = np.nanmin(np.asarray(self._obj.time))
            start_time = pd.to_datetime(start_time).strftime('%Y-%m-%d')
            end_time = np.nanmax(np.asarray(self._obj.time)) + \
                            np.timedelta64(23, 'h') + np.timedelta64(59, 'm')
            end_time = pd.to_datetime(end_time).strftime('%Y-%m-%d')
            times = pd.date_range(start_time, end_time, freq=freq)

        # Create empty dataset to hold interpolated values
        trajcoord = range(self._obj.dims['trajectory'])
        d = xr.Dataset(
            coords={
                'trajectory': (["trajectory"], trajcoord),
                'time': (["time"], times)
                    },
            attrs = self._obj.attrs
            )

        for varname, var in self._obj.variables.items():
            if varname in ['time', 'obs']:
                continue
            if self.timedim not in var.dims:
                d['varname'] = var
                continue

            # Create empty dataarray to hold interpolated values for given variable
            da = xr.DataArray(
                data=np.zeros(tuple(d.dims[di] for di in ['trajectory', 'time']))*np.nan,
                dims=d.dims,
                coords=d.coords,
                attrs=var.attrs
                )

            for t in range(self._obj.dims['trajectory']):  # loop over trajectories
                origtimes = self._obj['time'].isel(trajectory=t).astype(np.float64).values
                validtime = np.nonzero(~np.isnan(origtimes))[0]
                interptime = origtimes[validtime]
                interpvar = var.isel(trajectory=t).data
                # Make interpolator
                f = interp1d(interptime, interpvar, bounds_error=False)
                # Interpolate onto given times
                da.loc[{'trajectory': t}] = f(times.to_numpy().astype(np.float64))

            d[varname] = da

        return d
